/*file nlin.c
  revision log:
	19 Oct 1986 - made deriv, resid, wtj non static since we need them
			in the subroutine. - sel
	21 Aug 1986 - changed a, b, deriv, resid, and wtj to static
			variables - sel.
	5/21/85 - fix normalization of errors
  Steve Levine and Dick French
*/
#include <stdio.h>
#include <math.h>

#include "nldef.h"
				/*     RETURNED PARAMETERS :       */
extern double chisq;		/* wt'ed sum of squared residuals  */
extern double sigmay;		/* std. dev. per degree of freedom */
extern double cparam[];		/* current parameters              */
extern double sigpar[];		/* std. dev. of parameters         */
extern double cormat[][MAXPAR];	/* correlation matrix              */
extern int nitper;		/* number of iterations performed  */

/*	nlin returns an integer error flag, with the following meaning: 
 *	nerr =	1	ok                                             
 *		2	singular matrix when computing new parameters 
 *		3	didn't converge within nitmax iterations     
 *		4	# of data pts. < # of parameters            
 *		5	# of parameters < or = zero                
 *		6	nitmax < or = zero                        
 *		7	singular matrix when computing correlation matrix
 */

static double a[MAXPAR][MAXPAR];
static double b[MAXPAR];
double deriv[MAXPAR];

double resid = 0;
double wtj = 0;

nlin(subrtn,epsil,nitmax,numpar,numpts,itprnt)

/*	INPUT PARAMETERS					        
 *	subrtn	name of function to be called for computing model function
 *		calling sequence :
 *		subrtn(j, nitper)
 *			where j = index of data pt. to be processed
 *				  (j=0 for first data point)
 *			      nitper = current iteration number  
 *				  (nipter=1 for first iteration)
 *  		It is the responsibility of subrtn to compute the following:
 *			resid	(true-calc) for j'th point
 *			deriv[] array of derivatives of model function
 *				with respect to each parameter, for j'th pt.
 *			wtj	weight of j'th point
 *	epsil	convergence criterion, max. allowed fractional change in 
 *		any parameter between successive iterations		
 *	nitmax	max. number of iterations to be performed	
 *	numpar	number of parameters to be fit		
 *	numpts	number of data points		
 *	itprnt	flag: 0 = no printed output
 *		      1 = print chi-squared after each iteration
 *		      2 = and print current parameters	
 */
int (*subrtn)();
int numpar, numpts;
int nitmax, itprnt;
double epsil;
{
	int i, j, k, l, m;
	int nerr = 1;
	int cerr;
	int icon, nwted;
	double dmtx[MAXPAR][MAXPAR];
	double sqtvf,fabs(),sqrt();

/*
*	Check error conditions
*/
	if(numpts < numpar)
		nerr = 4;
	if(numpar <= 0)
		nerr = 5;
	if(nitmax <= 0)
		nerr = 6;
	if(nerr > 1)
		return(nerr);

	for(nitper = 1;nitper <= nitmax; nitper++)
	{
/*
*	Zero the accumulators
*/
		for(i = 0;i < numpar;i++)
		{
			b[i] = 0;
			for(j = 0;j < numpar; j++)
				a[i][j] = 0;
		}
/*
*	print parameters this iteration if desired
*/
		if(itprnt == 2)
		{
			printf("nlin: Iteration %d parameters\n",nitper);
			for(i = 0; i < numpar; i++)
				printf("cparam %d = %.16g\n", i, cparam[i]);
		}
/*
*	convert array to a square array
*/
		chisq = 0;
		for(j = 0; j < numpts; j++)
		{
			(*subrtn)(j, nitper);
			if(itprnt > 0)
				chisq = chisq + (resid * resid * wtj);
			for(l = 0; l < numpar; l++)
			{
				b[l] = b[l]+(deriv[l]*resid*wtj);
				for(m = 0; m < numpar; m++)
				{
					a[l][m]=a[l][m]+(deriv[l]*deriv[m]*wtj);
					dmtx[l][m] = a[l][m];
				}
			}
		}
/*		if(numpar == 1)		*/
/*
*	For >1 parameter
*/
		for(m = 1; m < numpar; m++)
		{
			for(i = 0; i < m - 1; i++)
			{
				a[m][i] = a[i][m];
				dmtx[m][i] = dmtx[i][m];
			}
		}
/*
*	read and compute corrections to previous values
*/
		nerr = clvlnr(numpar);
		if(nerr == 2)
			return(nerr);
/*
*	Update Parameters
*/
		for(i = 0; i < numpar; i++)
			b[i]  = cparam[i] + b[i];

/*
*	Test for convergence
*/
		icon = 0;
		for(i = 0; i < numpar; i++)
		{
			if(fabs((cparam[i]/b[i]) - 1.0) > epsil)
				icon = 1;
		}

/*
*	Set parameters for next iteration
*/
		for(i = 0; i < numpar; i++)
		{
			cparam[i] = b[i];
		}
		if(itprnt > 0)
			printf("Weighted sum of squared residuals = %.16g\n ",chisq);
/*
*	Drop out of loop if all parameters converged
*/
		if(icon == 0)
			goto parttwo;
	}
/*
*	no convergence
*/
	nerr = 3;
/*   01-23-88:  do all the rest of the stuff even though it did not
		converge within nitmax.
*/
	goto parttwo;
	return(nerr);
/*
*	calculate sum of squared residuals
*/
parttwo:
	chisq = 0;
	nwted = 0;
	for(j = 0;j < numpts; j++)
	{
		(*subrtn)(j, nitper);
		if(wtj != 0)
			nwted++;
		chisq = chisq + wtj * resid *resid;
	}
	sigmay = 0;
	if(nwted != numpar)
		sigmay = sqrt(chisq / (nwted - numpar));
/*
*	Calculate standard deviation and correlation matrix
*	inverting the solution matrix
*/
	for(i = 0; i < numpar; i++)
	{
		for(j = 0; j < numpar; j++)
			a[i][j] = dmtx[i][j];
	}
	for(i = 0; i < numpar; i++)
	{
		for(j = 0; j < numpar; j++)
			b[j] = 0;
		b[i] = 1;
		cerr = clvlnr(numpar);
		if(cerr == 2)
		{
			nerr = 7;
			return(nerr);
		}
		for(j = 0; j < numpar; j++)
		{
			cormat[i][j] = b[j];
			for(k = 0;k < numpar; k++)
				a[j][k] = dmtx[j][k];
		}
	}
/*
*	solve for sigma's
*/
	for(i = 0; i < numpar; i++)
		sigpar[i] = sqrt(cormat[i][i]);
/*
*	normalize the correlation matrix
*/
	for(i = 0; i < numpar; i++)
	{
		for(j = 0; j < numpar; j++)
			cormat[i][j] = cormat[i][j]/(sigpar[i]*sigpar[j]);
	}
/*
*	renormalize sigpar
*	-- note that it is assumed that the data are either zero or
*		have equal non-zero weight
*/
	for(i = 0; i < numpar; i++)
		sigpar[i] = sigpar[i] * sigmay;
	return(nerr);
}
	

clvlnr(numpar)
{

	int i, j, k;
	int irow, icolum, icl, irl, l;
	int iflag = 1;
	double pivsiz = 0;
	double ipivot[MAXPAR];
	double amax, aa, swap, pivot, t;
	double sqrt(),fabs(),pow();

	for(j = 0; j < numpar; j++)
		ipivot[j] = 0;
	for(i = 0; i < numpar; i++)
	{
		for(j = 0; j < numpar; j++)
			pivsiz = pivsiz + a[i][j] * a[i][j];
	}
	pivsiz = sqrt(pivsiz) * pow(10.,-16.)/(numpar*numpar);
	for(i = 0; i < numpar; i++)
	{
		amax = 0;
		for(j = 0;j < numpar; j++)
		{
			if(ipivot[j] != 1)
			{
				for(k = 0; k < numpar; k++)
				{
					if(ipivot[k] != 1)
					{	
						aa = fabs(a[j][k]);
						if(amax < aa)
						{
							irow = j;
							icolum = k;
							amax = aa;
						}
					}
				}
			}
		}
		ipivot[icolum]++;
		if(irow != icolum)
		{
			irl = irow;
			icl = icolum;
			for(l = 0; l < numpar; l++)
			{
				swap = a[icl][l];
				a[icl][l] = a[irl][l];
				a[irl][l] = swap;
			}
			swap = b[irow];
			b[irow] = b[icolum];
			b[icolum] = swap;
		}
		pivot = a[icolum][icolum];
		if(fabs(pivot) < pivsiz)
		{
			iflag = 2;
			return(iflag);
		}
		a[icolum][icolum] = 1;
		icl = icolum - numpar;
		for(l = 0; l < numpar; l++)
			a[icolum][l] = a[icolum][l]/pivot;
		b[icolum] = b[icolum]/pivot;
		for(j = 0;j < numpar; j++)
		{
			if(j != icolum)
			{
				t = a[j][icolum];
				a[j][icolum] = 0;
				for(k = 0; k < numpar; k++)
					a[j][k] = a[j][k] - a[icolum][k] * t;
			b[j] = b[j] - b[icolum] * t;
			}
		}
	}
	return(iflag);
}
