/*   simplex.c   */
/*   jjk:  adopted from Bill Press (see comment below)   */
/*   02-06-86   */

/*   NOTE:  NDIM must be set in simp.h and then library rebuilt */
/*   Actually, it is probably best for simplex.c and simp.h to be
	copied.  This can be just a safe place for them to be.  */

/*   This is a minimum finding routine known as simplex developed 
	by Nelder and Mead (Computer Journal, vol 7, p.308) that was
	given in a fortran program "Amoeba" written by Bill Press (?)
	in his book (p.292-3).  See Jack Wisdom for exact reference and
	comments of the original program.  

	The code is mine, based on the fortran program as found in
	the above reference; the names may be reminiscent of fortran
	styles since I kept nearly the same names for ease of
	transcription.  Also, the code is nearly the same as the
	fortran version, and if the logic or redundancy seems a
	little perverse, thank fortran (mostly).

	The calling sequence contains the following variables:
		
		p	matrix of linearly independant (non-degenerate)
			vectors (coordinates) which are the
			vertices of the starting simplex as the rows.
			At completion, p is set to a set of vectors,
			all of which are within the convergence
			criterion
			double p[ndim+1][ndim]

		y	initial values of func (see below) evaluated
			at vertices (rows) of p.
			y is set to the new points upon successful
			completion of the function
			double y[ndim+1]

		mp	ndim+1 (for redundancy's sake, I guess)
			int mp

		np	ndim (just for completeness!!)
			int np

		ndim	the number of dimensions of the function,
			func, to be minimized
			int ndim

		ftol	fractional convergence tolerance
			double ftol

		func	function of an n dimensional vector to be 
			minimized by this routine
			NOTE:  this function must be supplied!
			double func();

	The include file "~jjk/hy/initcond/simp.h" contains:
		
		NMAX	max number of dimensions
			02-11-86: no longer used, NDIM is used instead

		ALPHA	expansion parameter

		BETA	contraction parameter

		GAMMA	further expansion parameter

		ITMAX	Max number of iterations
	
		NDIM	number of dimensions actually necessary

	Upon successful completion, the program returns

		iter	number of iterations taken for success, returned
			int iter
	02-26-86:  add write to file "simpout"
	04-07-86:  check for reasonable offset
 */

#include "simp.h"
#include <math.h>
#include <stdio.h>

simplex(p, y, mp, np, ndim, ftol, func)
int mp, np, ndim;
double y[], p[][NDIM], ftol;
double (*func)();
{
    int i, j, lo, hi, nhi, mpts, iter;
    double a, v;
    double ypr, yprr;
    double pr[NDIM], prr[NDIM], pbar[NDIM];
    FILE *pout, *pout2, *wopen(), *rwclose();

    pout=wopen("simpout");
    mpts=ndim+1;
    iter=0;
    while(1)
    {
	/*
	 * 	Check to see if in reasonable bounds
	 */
	lo=0;
	if (y[0]>y[1])
	{
	    hi=0;
	    nhi=1;
	}
	else
	{
	    hi=1;
	    nhi=0;
	}
	for(i=0; i<mpts; i++)
	{
	    if(y[i]<y[lo])
		lo=i;
	    if(y[i]>y[hi])
	    {
		nhi=hi;
		hi=i;
	    }
	    else if(y[i]>y[nhi] && i!=hi)
		nhi=i;
	}
	if((2.*fabs(y[hi]-y[lo])/(fabs(y[hi])+fabs(y[lo])))<ftol ||
	    fabs(y[hi]-y[lo])<1e-14)
	{
	    rwclose(pout);
	    return(iter);
	}
	if(iter%10==0)
	{
	    rewind(pout);
	    for(i=0; i<mpts; i++)
	    {
	    	for(j=0; j<ndim; j++)
		{
	    	    fprintf(pout, "%20.16f ", p[i][j]);
	    	    if(j==ndim-1)
			 fprintf(pout, "\n");
		}
	    }
    	    fprintf(pout, "iter=%d, ftol=%e\n", iter, ftol);
    	    fprintf(pout, "y[%d]=%20.16e\n", lo, y[lo]);
    	    fprintf(pout, "y[%d]=%20.16e\n", hi, y[hi]);
	    fflush(pout);
	}
	if(iter++>=ITMAX)
	{
	    rwclose(pout);
	    printf("SIMPLEX:  exceeded max iterations\n");
	    return(iter);
	}
	for(j=0; j<ndim; j++)
	    pbar[j]=0;
	for(i=0; i<mpts; i++)
	    if(i != hi)
		for(j=0; j<ndim; j++)
		    pbar[j] += p[i][j];
	for(j=0; j<ndim; j++)
	{
	    pbar[j] *= 1./(double)ndim;
	    pr[j]=(1. + ALPHA)*pbar[j] - ALPHA*p[hi][j];
	}
	ypr=func(pr);
	if(ypr<=y[lo])
	{
	    for(j=0; j<ndim; j++)
		prr[j]=GAMMA*pr[j] + (1. - GAMMA)*pbar[j];
	    yprr=func(prr);
	    if(yprr<y[lo])
	    {
		for(j=0; j<ndim; j++)
		    p[hi][j]=prr[j];
		y[hi]=yprr;
	    }
	    else
	    {
		for(j=0; j<ndim; j++)
		    p[hi][j]=pr[j];
		y[hi]=ypr;
	    }
	}
	else if(ypr>=y[nhi])
	{
	    if(ypr<y[hi])
	    {
		for(j=0; j<ndim; j++)
		    p[hi][j]=pr[j];
		y[hi]=ypr;
	    }
	    for(j=0; j<ndim; j++)
		prr[j]=BETA*p[hi][j] + (1. - BETA)*pbar[j];
	    yprr=func(prr);
	    if(yprr<y[hi])
	    {
		for(j=0; j<ndim; j++)
		    p[hi][j]=prr[j];
		y[hi]=yprr;
	    }
	    else
		for(i=0; i<mpts; i++)
		    if(i != lo)
		    {
			for(j=0; j<ndim; j++)
			{
			    pr[j]=BETA*(p[i][j] + p[lo][j]);
			    p[i][j]=pr[j];
			}
			y[i]=func(pr);
		    }
	}
	else
	{
	    for(j=0; j<ndim; j++)
		p[hi][j]=pr[j];
	    y[hi]=ypr;
	}
    }
}
