/*---------------------------------------------------------------
           FILE: limbeval.c
      $Revision: 1.1 $
         $State: Exp $
  Last submitted $Date: 1995/04/14 01:55:57 $ $Author: jh $

   Locked out by $Locker: jh $

  REVISION HISTORY
  $Log: limbeval.c,v $
 * Revision 1.1  1995/04/14  01:55:57  jh
 * Initial revision
 *

  $Source: /home/jh/ast/jiggle/limbctr/RCS/limbeval.c,v $

  Copyright (c) 1993 Joseph Harrington.  All Rights reserved.
---------------------------------------------------------------*/
#ifndef LINT
static char rcsid[] = "$Header: /home/jh/ast/jiggle/limbctr/RCS/limbeval.c,v 1.1 1995/04/14 01:55:57 jh Exp jh $";
static char copyright[] =
    "Copyright (c) 1993 Joseph Harrington.  All Rights reserved.";
#endif
/*---------------------------------------------------------------
limbeval.c = quality evaluation functions for fitting ellipses to
images

External object files or libraries:
bilin_interp.c
---------------------------------------------------------------*/
/* inclusions */
#include <math.h>
#include "bilin_interp.h"
#include "limbeval.h"
#include "planetmap.h"

/* definitions */


/*---------------------------------------------------------------
Effect: Evaluates the quality fit of an ellipse p to an image of
	a planet represented by limb, and returns that value.

Requires: p: array of double-precision ellipse parameters:
		LE_xc(p) = x center
		LE_yc(p) = y center
		LE_re(p) = semimajor axis
		LE_lt(p) = limb thickness parameter, if fit

	  limb: pointer to Limbevalpars for this image and limb. 

Modifies: nothing.
Implementation notes:
---------------------------------------------------------------*/
double
limbeval(
	 double        p[],
	 Limbevalpars *limb
){
    double  xc = LE_xc(p);
    double  yc = LE_yc(p);
    double  re = LE_re(p);
    double  lt;
    double  ellip  = limb->ellip;
    double  rc = limb->rc;
    double  rs = limb->rs;
    double  dcircang = 2 * PI / limb->npoint;
    Doub2  *im = limb->image;
    double  xe;
    double  ye;
    double  x;
    double  y;
    double  circang;
    double  pdlat;
    double  pixel;
    double  qual = 0.;

    unsigned int count = 0;

    /* check reasonability limits */
    if ( fabs(re - limb->nom_re) > limb->delt_re || re < 0. ){
	return DBL_MAX;
    }

    /* step around the ring (slightly oversample pole, but easy to compute) */
    for (circang = 0.; circang < 2. * PI ; circang += dcircang){

	/* coordinates in the unrotated ellipse (speed hack: tabulate trig) */
	xe =         re * cos(circang);
	ye = ellip * re * sin(circang);

	/* limb thickness parameter */
	if(limb->fit_lt){
	    lt = LE_lt(p);
	    if(	fabs(lt) > limb->delt_lt ){
		return DBL_MAX;
	    }
	    pdlat = PMpdfca(circang, ellip);
	    xe += lt * cos(pdlat);
	    ye += lt * sin(pdlat);
	}

	/* rotate and shift to center of frame */
	x = xc + xe *  rc + ye * rs;
	y = yc + xe * -rs + ye * rc;

	/* get pixel value, add to evaluation if it's in bounds */
	pixel = bilin_interp(im, y, x);
	if (pixel < DBL_MAX){
	    qual += pixel;
	    count++;
	}
    }

    /* check to see we got enough ellipse */
    if (count == 0 || count < limb->minfrac * limb->npoint){
	return DBL_MAX;
    }

    return qual/(double)count;
}

/*------------------------create_limbevalpars----------------------
Effect: Allocates and initializes a Limbevalpars.  Returns a
	pointer to the Limbevalpars or NULL if an error occurs.

Requires: nothing.

Modifies: nothing.

Implementation notes: free returned Limbevalpars with
	free_limbevalpars().
---------------------------------------------------------------*/
Limbevalpars *
create_limbevalpars(void){
    Limbevalpars *limb;
    
    limb = (Limbevalpars *)malloc(sizeof(*limb));
    if (!limb){
	fprintf(stderr, "create_limbevalpars: unable to allocate %u bytes\n",
		sizeof(*limb));
	return NULL;
    }

    /* defaults */
    limb->imagename[0] = '\0';	/* name of image to fit to */
    limb->image = NULL;		/* image to fit to */
    limb->nom_re = DBL_MAX;	/* nominal equatorial radius */
    limb->delt_re = DBL_MAX;	/* max. deviation from re */
    limb->delt_lt = 0.;		/* max. limb thickness */
    limb->fit_lt = FALSE;       /* fit limb thickness? */
    limb->fit_npars = LE_NLTNP; /* number of fit parameters */
    limb->ellip = 1.;		/* ellipticity of planetary disc */
    limb->rc = 1.;		/* cosine of coordinate rotation angle */
    limb->rs = 0.;		/* sine of coordinate rotation angle */
    limb->npoint = 360;		/* number of points to evaluate */
    limb->minfrac = 1.;		/* minimum fraction of ellipse on image */
    limb->nom_xc = DBL_MAX;	/* X of nominal planet center */
    limb->nom_yc = DBL_MAX;	/* Y of nominal planet center */

    return limb;
}

/*------------------------free_limbevalpars------------------------
Effect: Frees a Limbevalpars, including the image.

Requires: limb: pointer to Limbevalpars to free.

Modifies: *limb.

Implementation notes:
---------------------------------------------------------------*/
void
free_limbevalpars(Limbevalpars *limb){
    free_doub2(limb->image);
    free(limb);

    return;
}
