/*---------------------------------------------------------------
           FILE: memory.c
      $Revision: 2.0 $
         $State: Exp $
  Last submitted $Date: 93/03/30 18:04:42 $ $Author: jh $

   Locked out by $Locker: jh $

  REVISION HISTORY
  $Log:	memory.c,v $
 * Revision 2.0  93/03/30  18:04:42  jh
 * Actual first release version.
 * 
 * Revision 1.2  93/03/30  13:54:20  jh
 * General code cleanup, testing, and added features.
 * First release version.
 * 
 * Revision 1.1  92/12/15  13:07:14  jh
 * Initial revision
 * 

  $Source: /u3/jh/jupobs92/anal/iraf/src/jiggle/RCS/memory.c,v $

  Copyright (c) 1992, 1993 Joseph Harrington.
  All Rights reserved.
---------------------------------------------------------------*/
#ifndef LINT
static char rcsid[] = "$Header: /u3/jh/jupobs92/anal/iraf/src/jiggle/RCS/memory.c,v 2.0 93/03/30 18:04:42 jh Exp Locker: jh $";
static char copyright[] =
    "Copyright (c) 1992, 1993 Joseph Harrington.  All Rights reserved.";
#endif
/*---------------------------------------------------------------
memory.c = functions to allocate arrays
---------------------------------------------------------------*/
/* definitions */

#define TRUE  1
#define FALSE 0

/* inclusions */

#include <stdio.h>
#include <stdlib.h>
#include "memory.h"

/*---------------------alloc_doub2-------------------------------
Effect: Creates a Doub2 and initializes it.  Returns a pointer
        to it or NULL.

Requires: ni, nj: dimensions of the array; nj varies fastest.

Modifies: nothing.

Implementation notes: Doub2 should be freed with free_doub2().
---------------------------------------------------------------*/
Doub2 *
alloc_doub2(
	    size_t ni,
	    size_t nj
){
    Doub2 *m;

    /* mallocate some memory */

    m = (Doub2 *)malloc(sizeof(*m));
    if (!m) {
	fprintf(stderr, "alloc_doub2: unable to allocate %u bytes\n",
		sizeof(*m));
	return NULL;
    }

    m->data = alloc_arrdoub2(ni, nj);
    if (!m->data) {
	free(m);
	return NULL;
    }

    /* initialize it */
    m->ni = ni;
    m->nj = nj;

    /* return it */
    return m;
}

/*----------------------alloc_arrdoub2---------------------------
Effect: Allocates and organizes memory for a 2D array of
        doubles.  Returns pointer to array or NULL if a malloc
	error occurred.

Requires: ni, nj: dimensions of the array; nj varies fastest.

Modifies: nothing.

Implementation notes: Array should be freed with free_arrdoub2().
---------------------------------------------------------------*/
double **
alloc_arrdoub2(
	       size_t ni,
	       size_t nj
){
    size_t   i;
    double **m;
    
    /* mallocate some memory */

    m = (double **)malloc(ni * sizeof(*m));
    if (!m) {
	fprintf(stderr, "alloc_arrdoub2: unable to allocate %u bytes\n",
		ni * sizeof(*m));
	return NULL;
    }

    m[0] = (double *)malloc(ni * nj * sizeof(*m[0]));
    if (!m[0]) {
	fprintf(stderr, "alloc_arrdoub2: unable to allocate %u bytes\n",
		ni * nj * sizeof(*m[0]));
	free(m);
	return NULL;
    }

    /* set up pointers to array rows */

    for (i = 1 ; i < ni ; i++) {
	m[i] = m[0] + i * nj;
    }

    return m;
}

/*--------------------------free_arrdoub2-------------------------
Effect: frees memory allocated by alloc_arrdoub2().

Requires: m: 2d array created by alloc_arrdoub2().

Modifies: *m

Implementation notes:
---------------------------------------------------------------*/
void
free_arrdoub2(double **m){
    if ( m == NULL ){
	return;
    }
    free(m[0]);
    free(m);
    return;
}

/*--------------------------free_doub2-----------------------------
Effect: frees memory allocated by alloc_doub2().

Requires: m: Doub2 created by alloc_doub2().

Modifies: *m

Implementation notes:
---------------------------------------------------------------*/
void
free_doub2(Doub2 *m){
    if ( m == NULL ){
	return;
    }
    free_arrdoub2(m->data);
    free(m);
    return;
}

/*---------------------alloc_size_t2------------------------------
Effect: Creates a Size_T2 and initializes it.  Returns a pointer
        to it or NULL.

Requires: ni, nj: dimensions of the array; nj varies fastest.

Modifies: nothing.

Implementation notes: Size_T2 should be freed with
		      free_size_t2().
---------------------------------------------------------------*/
Size_T2 *
alloc_size_t2(
	      size_t ni,
	      size_t nj
){
    Size_T2 *m;

    /* mallocate some memory */

    m = (Size_T2 *)malloc(sizeof(*m));
    if (!m) {
	fprintf(stderr, "alloc_size_t2: unable to allocate %u bytes\n",
		sizeof(*m));
	return NULL;
    }

    m->data = alloc_arrsize_t2(ni, nj);
    if (!m->data) {
	free(m);
	return NULL;
    }

    /* initialize it */
    m->ni = ni;
    m->nj = nj;

    /* return it */
    return m;
}

/*----------------------alloc_arrsize_t2---------------------------
Effect: Allocates and organizes memory for a 2D array of
        size_ts.  Returns pointer to array or NULL if a malloc
	error occurred.

Requires: ni, nj: dimensions of the array; nj varies fastest.

Modifies: nothing.

Implementation notes: Array should be freed with
                      free_arrsize_t2().
---------------------------------------------------------------*/
size_t **
alloc_arrsize_t2(
		 size_t ni,
		 size_t nj
){
    size_t   i;
    size_t **m;
    
    /* mallocate some memory */

    m = (size_t **)malloc(ni * sizeof(*m));
    if (!m) {
	fprintf(stderr, "alloc_arrsize_t2: unable to allocate %u bytes\n",
		ni * sizeof(*m));
	return NULL;
    }

    m[0] = (size_t *)malloc(ni * nj * sizeof(*m[0]));
    if (!m[0]) {
	fprintf(stderr, "alloc_arrsize_t2: unable to allocate %u bytes\n",
		ni * nj * sizeof(*m[0]));
	free(m);
	return NULL;
    }

    /* set up pointers to array rows */

    for (i = 1 ; i < ni ; i++) {
	m[i] = m[0] + i * nj;
    }

    return m;
}

/*--------------------------free_arrsize_t2-------------------------
Effect: frees memory allocated by alloc_arrsize_t2().

Requires: m: 2d array created by alloc_arrsize_t2().

Modifies: *m

Implementation notes:
---------------------------------------------------------------*/
void
free_arrsize_t2(size_t **m){
    if ( m == NULL ){
	return;
    }
    free(m[0]);
    free(m);
}

/*--------------------------free_size_t2-----------------------------
Effect: frees memory allocated by alloc_size_t2().

Requires: m: Size_T2 created by alloc_size_t2().

Modifies: *m

Implementation notes:
---------------------------------------------------------------*/
void
free_size_t2(Size_T2 *m){
    if ( m == NULL ){
	return;
    }
    free_arrsize_t2(m->data);
    free(m);
    return;
}

/*---------------------alloc_doub2p2-----------------------------
Effect: Creates a Doub2p2 and initializes it.  Returns a pointer
        to it or NULL.  Does not create the Doub2s it points to.

Requires: ni, nj: dimensions of the array; nj varies fastest.

Modifies: nothing.

Implementation notes: Doub2p2 should be freed with
		      free_doub2p2().
---------------------------------------------------------------*/
Doub2p2 *
alloc_doub2p2(
	      size_t ni,
	      size_t nj
){
    Doub2p2 *m;

    /* mallocate some memory */

    m = (Doub2p2 *)malloc(sizeof(*m));
    if (!m) {
	fprintf(stderr, "alloc_doub2p2: unable to allocate %u bytes\n",
		sizeof(*m));
	return NULL;
    }

    m->data = alloc_arrdoub2p2(ni, nj);
    if (!m->data) {
	free(m);
	return NULL;
    }

    /* initialize it */
    m->ni = ni;
    m->nj = nj;

    /* return it */
    return m;
}

/*----------------------alloc_arrdoub2p2------------------------
Effect: Allocates and organizes memory for a 2D array of
        Doub2 *s.  Returns pointer to array or NULL if a malloc
	error occurred.

Requires: ni, nj: dimensions of the array; nj varies fastest.

Modifies: nothing.

Implementation notes: Array should be freed with
                      free_arrdoub2p2().
---------------------------------------------------------------*/
Doub2 ***
alloc_arrdoub2p2(
		 size_t ni,
		 size_t nj
){
    size_t   i;
    size_t   j;
    Doub2 ***m;
    
    /* mallocate some memory */

    m = (Doub2 ***)malloc(ni * sizeof(*m));
    if (!m) {
	fprintf(stderr, "alloc_arrdoub2p2: unable to allocate %u bytes\n",
		ni * sizeof(*m));
	return NULL;
    }

    m[0] = (Doub2 * *)malloc(ni * nj * sizeof(*m[0]));
    if (!m[0]) {
	fprintf(stderr, "alloc_arrdoub2p2: unable to allocate %u bytes\n",
		ni * nj * sizeof(*m[0]));
	free(m);
	return NULL;
    }

    /* set up pointers to array rows */

    for (i = 1 ; i < ni ; i++) {
	m[i] = m[0] + i * nj;
    }

    /* initialize to NULL so free_arrdoub2p2elts can deal */

    for (i = 0 ; i < ni ; i++) {
	for (j = 0 ; j < nj ; j++) {
	    m[i][j] = NULL;
	}
    }

    return m;
}

/*--------------------------free_arrdoub2p2----------------------
Effect: frees memory allocated by alloc_arrdoub2p2().

Requires: m: 2d array created by alloc_arrdoub2p2().

Modifies: *m

Implementation notes: 
---------------------------------------------------------------*/
void
free_arrdoub2p2(Doub2 ***m){
    if ( m == NULL ){
	return;
    }
    free(m[0]);
    free(m);
    return;
}

/*--------------------------free_doub2p2-------------------------
Effect: frees memory allocated by alloc_doub2p2().

Requires: m: Doub2p2 created by alloc_doub2p2().

Modifies: *m

Implementation notes:
---------------------------------------------------------------*/
void
free_doub2p2(Doub2p2 *m){
    if ( m == NULL ){
	return;
    }
    free_arrdoub2p2(m->data);
    free(m);
    return;
}

/*--------------------------alloc_arrdoub2p2elts-----------------
Effect: Allocates Doub2's and points a 2D array of Doub2 *s at
        them.  Returns SUCCESS or FAILURE as appropriate.

Requires: m: 2D array of Doub2's created by alloc_doub2p2().
          ni, nj: dimensions of the array; nj varies fastest.
          ny, nx: dimensions of the Doub2's; nx varies fastest.

Modifies: *m

Implementation notes: Free with free_arrdoub2p2elts before
		      calling free_doub2p2(). 
---------------------------------------------------------------*/
int
alloc_arrdoub2p2elts(
		 Doub2 ***m,
		 size_t   ni,
		 size_t   nj,
		 size_t   ny,
		 size_t   nx
){
    size_t i;
    size_t j;

    for ( i = 0 ; i < ni ; i++ ){
	for ( j = 0 ; j < nj ; j++ ){
	    m[i][j] = alloc_doub2(ny, nx);
	    if (!m[i][j]){
		free_arrdoub2p2elts(m, ni, nj);
		return FAILURE;
	    }
	}
    }

    return SUCCESS;
}

/*--------------------------free_arrdoub2p2elts---------------------
Effect: frees Doub2's pointed to by a 2D array of Doub2 *s.

Requires: m: 2D array of Doub2's created by alloc_doub2p2().
          ni, nj: dimensions of the array; nj varies fastest.

Modifies: *m

Implementation notes: Call this before calling free_doub2p2().
---------------------------------------------------------------*/
void
free_arrdoub2p2elts(
		 Doub2 ***m,
		 size_t   ni,
		 size_t   nj
){
    size_t i;
    size_t j;

    if ( m == NULL ){
	return;
    }
    for ( i = 0 ; i < ni ; i++ ){
	for ( j = 0 ; j < nj ; j++ ){
	    if (!m[i][j]){
		free_doub2(m[i][j]);
	    }
	}
    }
    return;
}

/*---------------------alloc_char3-------------------------------
Effect: Creates a Char3 and initializes it.  Returns a pointer
        to it or NULL.

Requires: ni, nj, nk: dimensions of the array; nk varies fastest,
                      then nj.

Modifies: nothing.

Implementation notes: Char3 should be freed with free_char3().
---------------------------------------------------------------*/
Char3 *
alloc_char3(
	    size_t ni,
	    size_t nj,
	    size_t nk
){
    Char3 *m;

    /* mallocate some memory */

    m = (Char3 *)malloc(sizeof(*m));
    if (!m) {
	fprintf(stderr, "alloc_char3: unable to allocate %u bytes\n",
		sizeof(*m));
	return NULL;
    }

    m->data = alloc_arrchar3(ni, nj, nk);
    if (!m->data) {
	free(m);
	return NULL;
    }

    /* initialize it */
    m->ni = ni;
    m->nj = nj;
    m->nk = nk;

    /* return it */
    return m;
}

/*----------------------alloc_arrchar3---------------------------
Effect: Allocates and organizes memory for a 3D array of
        chars.  Returns pointer to array or NULL if a malloc
	error occurred.

Requires: ni, nj, nk: dimensions of the array; nk varies fastest,
                      then nj.
Modifies: nothing.

Implementation notes: Array should be freed with free_arrchar3().
---------------------------------------------------------------*/
char ***
alloc_arrchar3(
	       size_t ni,
	       size_t nj,
	       size_t nk
){
    size_t   i;
    size_t   j;
    char ***m;
    
    /* mallocate some memory */

    m = (char ***)malloc(ni * sizeof(*m));
    if (!m) {
	fprintf(stderr, "alloc_arrchar3: unable to allocate %u bytes\n",
		ni * sizeof(*m));
	return NULL;
    }

    m[0] = (char **)malloc(ni * nj * sizeof(*m[0]));
    if (!m[0]) {
	fprintf(stderr, "alloc_arrchar3: unable to allocate %u bytes\n",
		ni * nj * sizeof(*m[0]));
	free(m);
	return NULL;
    }

    m[0][0] = (char *)malloc(ni * nj * nk * sizeof(*m[0][0]));
    if (!m[0][0]) {
	fprintf(stderr, "alloc_arrchar3: unable to allocate %u bytes\n",
		ni * nj * nk * sizeof(*m[0][0]));
	free(m[0]);
	free(m);
	return NULL;
    }

    /* set up pointers to array sheets and rows */

    for ( i = 0 ; i < ni ; i++ ) {
	if ( i != 0 ){
	    m[i] = m[0] + i * nj;
	}
	for ( j = 0 ; j < nj ; j++ ) {
	    if ( (i != 0) || (j != 0) ){
		m[i][j] = m[0][0] + i * nj * nk + j * nk;
	    }
	}
    }

    return m;
}

/*--------------------------free_arrchar3-------------------------
Effect: frees memory allocated by alloc_arrchar3().

Requires: m: 2d array created by alloc_arrchar3().

Modifies: *m

Implementation notes:
---------------------------------------------------------------*/
void
free_arrchar3(char ***m){
    if ( m == NULL ){
	return;
    }
    free(m[0][0]);
    free(m[0]);
    free(m);
    return;
}

/*--------------------------free_char3-----------------------------
Effect: frees memory allocated by alloc_char3().

Requires: m: Char3 created by alloc_char3().

Modifies: *m

Implementation notes:
---------------------------------------------------------------*/
void
free_char3(Char3 *m){
    if ( m == NULL ){
	return;
    }
    free_arrchar3(m->data);
    free(m);
    return;
}
