/*---------------------------------------------------------------
           FILE: simplex.c
      $Revision: 2.0 $
         $State: Exp $
  Last submitted $Date: 93/03/30 18:04:56 $ $Author: jh $

   Locked out by $Locker: jh $

  REVISION HISTORY
  $Log:	simplex.c,v $
 * Revision 2.0  93/03/30  18:04:56  jh
 * Actual first release version.
 * 
 * Revision 1.2  93/03/30  13:54:23  jh
 * General code cleanup, testing, and added features.
 * First release version.
 * 
 * Revision 1.1  92/12/15  13:10:48  jh
 * Initial revision
 * 

  $Source: /u3/jh/jupobs92/anal/iraf/src/jiggle/RCS/simplex.c,v $

  Copyright (c) 1992, 1993 Joseph Harrington.
  All Rights reserved.
---------------------------------------------------------------*/
#ifndef LINT
static char rcsid[] = "$Header: /u3/jh/jupobs92/anal/iraf/src/jiggle/RCS/simplex.c,v 2.0 93/03/30 18:04:56 jh Exp Locker: jh $";
static char copyright[] =
    "Copyright (c) 1992, 1993 Joseph Harrington.  All Rights reserved.";
#endif
/*---------------------------------------------------------------
simplex.c = functions for simplex method of N-dimensional
            function minimization.  See Numerical Recipes
	    section 10.4.

External object files or libraries:
memory.c
-lm
---------------------------------------------------------------*/
/* inclusions */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "memory.h"
#include "simplex.h"

/*-------------------------amoeba-------------------------------
Effect: Simulates the form of the Numerical Recipes C function.

Requires: p:	2d array of doubles which are the coordinates of
                each vertex in the simplex, coords vary fastest.

	  y:    array of doubles, the values of the function at
                each point.

          ndim: size_t number of dimensions in the space.

          ftol: double fractional termination tolerance.

          funk: pointer to function returning double that is to
	        be minimized.

          nfunk: pointer to unsigned integer that will be set to
	         the number of calls to *funk.

Modifies: nfunk

Implementation notes:
---------------------------------------------------------------*/
void
amoeba(
       double **p,
       double y[],
       size_t ndim,
       double ftol,
       double (*funk)(),
       unsigned int *nfunk
){
    Simplex_param *param;
    size_t i;     /* counter over simplex points */
    size_t j;     /* counter over dimensions */

    /* get a param */
    param = simplex_create(ndim);
    if (!param){
	/* no error message because other functions will give them */
	*nfunk = 0;
	return;
    }

    /* put in the numbers from Numerical Recipes */
    /* note that NR doesn't let you pass parameters to funk */
    (void)simplex_init_param(param, funk, NULL, ftol, 1., 2., 0.5, 5000);

    /* copy in p and y */
    (void)simplex_init_simplex(param, p, y);

    /* do it */
    (void)simplex(param);
    *nfunk = param->nfunc;

    /* replace the initial simplex and function values with the found ones */
    for ( j = 0 ; j < param->npts ; j++ ){
	for ( i = 0 ; i < param->ndim ; i++ ){
	    p[j][i] = param->p[j][i];
	}
	y[j] = param->y[j];
    }

    /* clean up */
    simplex_free(param);

    return;
}

/*-------------------------simplex-------------------------------
Effect: Finds the local minimum of a function.

Requires: param: Simplex_param loaded and initialized; see
                 simplex.h and Numerical Recipes 10.4.

Modifies: *param->p, *param->y, param->nfunc

Implementation notes:
---------------------------------------------------------------*/
Simplex_param *
simplex(Simplex_param *param){
    size_t i;     /* counter over simplex points */
    size_t j;     /* counter over dimensions */
    size_t hi;    /* index of highest value of param->y */
    size_t nhi;   /* index of next highest value of param->y */
    size_t lo;    /* index of lowest value of param->y */
    double rtol;  /* fractional range in param->y */
    double ytry;  /* trial function evaluation */
    double ysave; /* previous high value */

    while(1){
	/* find lowest, highest, and next highest points in simplex */
	for ( i = 1, lo = 0 ; i < param->npts ; i++ ){
	    if ( param->y[i] < param->y[lo] ){
		lo = i;
	    }
	}

	if ( param->y[0] > param->y[1] ){
	    hi = 0;
	    nhi = 1;
	} else {
	    hi = 1;
	    nhi = 0;
	}

	for ( i = 2 ; i < param->npts ; i++ ){
	    if ( param->y[i] > param->y[hi] ){
		nhi = hi;
		hi = i;
	    } else if ( param->y[i] > param->y[nhi] ){
		nhi = i;
	    }
	}

	/* Termination Conditions */
	/* Compute fractional range and return if within ftol */
	/* fabs in denominator to deal with the case of */
	/* param->y[hi] and param->y[lo] being opposite in sign. */
	/* Denominator has 1. added (different from Numerical Recipes) */
	/* to keep it from blowing up if the value near the minimum is 0. */

	rtol = fabs( (param->y[hi] - param->y[lo]) / 
		     (0.5 * (  fabs(param->y[hi])
			     + fabs(param->y[lo]) 
			     + 1. ))
		   );

#ifdef DEBUG
	fprintf(stderr, "vert value         hi  coords\n");
	for ( i = 0 ; i < param->npts ; i++ ){
	    fprintf(stderr, "%4d % e ", i, param->y[i]);
	    if ( i == lo ) {
		fprintf(stderr, "lo  ");
	    } else if ( i == nhi ) {
		fprintf(stderr, "nhi ");
	    } else if ( i == hi ) {
		fprintf(stderr, "hi  ");
	    }
	    for ( j = 0 ; j < param->ndim ; j++ ){
		fprintf(stderr, "% e ", param->p[i][j]);
	    }
	    fprintf(stderr, "\n");
	}
	fprintf(stderr, "\n");
	fprintf(stderr, 
		"rtol: %e ftol: %e rtol/ftol: %e nfunc: %d\n",
		rtol, param->ftol, rtol/param->ftol, param->nfunc);
	fprintf(stderr, "\n");
#endif /* DEBUG */

	if ( rtol < param->ftol ){
	    break;
	}

	if ( param->nfunc >= param->funcmax ){
	    fprintf(stderr,
		    "simplex: exceeding %d evaluations.  rtol/ftol = %e.\n",
		    param->funcmax, rtol/param->ftol);
	    break;
	}

	/* New Trials */
	ytry = simplex_project(param, hi, param->flip);
	if ( ytry <= param->y[lo] ){
#ifdef DEBUG
	    fprintf(stderr, "flip hi, try to extend\n");
#endif /* DEBUG */		
	    ytry = simplex_project(param, hi, -param->extend);
	} else if ( ytry >= param->y[nhi] ){
	    ysave = param->y[hi];
	    ytry = simplex_project(param, hi, -param->contract);
	    if ( ytry >= ysave ){
#ifdef DEBUG
		fprintf(stderr, "shrink around lo\n");
#endif /* DEBUG */		
		for ( i = 0 ; i < param->npts ; i++) {
		    if ( i != lo ) {
			for ( j = 0 ; j < param->ndim ; j++ ){
			    param->p[i][j] = 0.5 * (  param->p[i][j]
						    + param->p[lo][j] );
			}
			param->y[i] =
			    (*param->func)(param->p[i], param->funcpars);
		    }
		}
		param->nfunc += param->ndim;
		simplex_calc_psum(param);
	    }
#ifdef DEBUG
	    else {
		fprintf(stderr, "contract hi\n");
	    }
#endif /* DEBUG */		
	}
#ifdef DEBUG
	else {
	    fprintf(stderr, "flip hi\n");
	}
#endif /* DEBUG */		
    }
    return param;
}

/*--------------------simplex_calc_psum-------------------------
Effect: (Re)calculates psum.

Requires:
	param: Pointer to Simplex_param being worked on.

Modifies: param.

Implementation notes: internal only.
---------------------------------------------------------------*/
void
simplex_calc_psum(Simplex_param *param){
    size_t i;
    size_t j;

    for ( j = 0 ; j < param->ndim ; j++ ){
	for ( i = 0 , param->psum[j] = 0. ; i < param->npts ; i++ ){
	    param->psum[j] += param->p[i][j];
	}
    }
    return;
}

/*--------------------simplex_project----------------------------
Effect: Projects simplex point indexed vert through the center of
        the opposing face by factor times the distance between
	vert and face from that face.  Evaluates *param->func
	there, and replaces indexed point with this new point & 
	its value if function evaluates lower there.  Adjusts
	psum, increments nfunc.  Returns final value at vert.

Requires:
	param: Pointer to Simplex_param being worked on.

	vert: Index of vertex on projection line.

	factor: Amount to project by.

Modifies: param.

Implementation notes: internal only.
---------------------------------------------------------------*/
double
simplex_project(
		Simplex_param *param,
		size_t         vert,
		double         factor
){
    size_t  j;
    double  fac1;
    double  fac2;
    double  ytry;

    /* calculate projection factors */

    fac1 = ( 1. + factor ) / (double)param->ndim;
    fac2 = fac1 + factor;

    /* calculate projection */

    for ( j = 0 ; j < param->ndim ; j++ ){
	param->ptry[j] = fac1 * param->psum[j] - fac2 * param->p[vert][j];
    }

    /* evaluate and count */

    ytry = (*param->func)( param->ptry, param->funcpars);
    param->nfunc++;

    /* if it's better than vert, replace vert */

    if ( ytry < param->y[vert] ){
	param->y[vert] = ytry;
	for ( j = 0 ; j < param->ndim ; j++ ){
	    param->psum[j] += param->ptry[j] - param->p[vert][j];
	    param->p[vert][j] = param->ptry[j];
	}
    }

    return ytry;
}

/*--------------------simplex_create----------------------------
Effect: Allocates a Simplex_param.  Returns NULL if a malloc
        error occurs.

Requires:
	ndim: number of parameters (dimensions) of func.

Modifies: Nothing.

Implementation notes:
---------------------------------------------------------------*/
Simplex_param *
simplex_create(size_t         ndim){
    Simplex_param *param;

    /* mallocate some memory */

    param = (Simplex_param *) malloc(sizeof(*param));
    if(!param){
	fprintf(stderr, "simplex_create: unable to allocate %u bytes.\n",
		sizeof(*param));
	return NULL;
    }

    /* set parameters here, we'll need some */

    param->ndim  = ndim;
    param->npts  = ndim + 1;
    param->nfunc = 0;

    /* mallocate the rest */

    param->y = (double *) malloc(param->npts * sizeof(*param->y));
    if(!param->y){
	fprintf(stderr, "simplex_create: unable to allocate %u bytes.\n",
		param->npts * sizeof(*param->y));
	free(param);
	return NULL;
    }

    param->p = alloc_arrdoub2(param->npts, ndim);
    if(!param->p){
	/* alloc_arrdoub2 produces an error message */
	free(param->y);
	free(param);
	return NULL;
    }

    param->psum = (double *)malloc(param->ndim * sizeof(*param->psum));
    if (!param->psum){
	fprintf(stderr, "simplex_create: unable to allocate %u bytes.\n",
		param->ndim * sizeof(*param->psum));
	free_arrdoub2(param->p);
	free(param->y);
	free(param);
	return NULL;
    }

    param->ptry = (double *)malloc(param->ndim * sizeof(*param->ptry));
    if (!param->ptry){
	fprintf(stderr, "simplex_create: unable to allocate %u bytes.\n",
		param->ndim * sizeof(*param->ptry));
	free(param->psum);
	free_arrdoub2(param->p);
	free(param->y);
	free(param);
	return NULL;
    }

    return param;
}

/*-----------------------simplex_init_param--------------------
Effect: Initializes constant parts of a Simplex_param.  Returns
        param.

Requires:
        param: pointer to Simplex_param with ndim and npts set.

        func: pointer to a function that takes an array of ndim
	      doubles and a void pointer (which can be used to
	      pass parameters) and returns a double.  This is
	      the function we're going to minimize.

	funcpars: void pointer that can be set to parameters
	      for func, or NULL. 

	ftol: find a minimum within this amount of a real
	      minimum.  Should not be less than machine
	      precision.

	flip: when inverting the simplex, fraction of current
	      height to place the vertex on the other side of the
	      base.

	extend: amount to extend the simplex by.

	contract: amount to contract the simplex by.

	funcmax: maximum number of function evaluations.

Modifies: param

Implementation notes:
---------------------------------------------------------------*/
Simplex_param *
simplex_init_param(
	     Simplex_param *param,
	     double       (*func)(),
	     void          *funcpars,
	     double         ftol,
	     double         flip,
	     double         extend,
	     double         contract,
	     unsigned int   funcmax
){
    param->func     = func;
    param->funcpars = funcpars;
    param->ftol     = ftol;
    param->flip     = flip;
    param->extend   = extend;
    param->contract = contract;
    param->funcmax  = funcmax;

    return param;
}

/*-----------------------simplex_init_p-------------------------
Effect: Copies p into param->p[0].  Returns param.

Requires:
        param: pointer to Simplex_param.

	p: array of param->ndim doubles which is an initial
	   vertex for the simplex.

Modifies: param.

Implementation notes:
---------------------------------------------------------------*/
Simplex_param *
simplex_init_p(
		   Simplex_param *param,
		   double        *p
){
    size_t i;

    for ( i = 0 ; i < param->ndim ; i++){
	param->p[0][i] = p[i];
    }

    return param;
}

/*-----------------------simplex_init_point---------------------
Effect: Generates an initial simplex with one vertex at p and
        with sides of length side.  Evaluates param->psum.
	Initializes param->y by evaluating *param->func.
	Returns param.

Requires:
        param: pointer to Simplex_param with all parameters but
	       p and y set, and with p[0] set.

	side: the distance between param->p[0] and the other
	      vertices of the simplex.

Modifies: param.

Implementation notes: separate from simplex_init_p for use in
                      restarts.
---------------------------------------------------------------*/
Simplex_param *
simplex_init_point(
	     Simplex_param *param,
	     double         side
){
    size_t i;     /* counter over simplex points */
    size_t j;     /* counter over dimensions */

    for ( i = 0 ; i < param->npts ; i++){
	for ( j = 0 ; j < param->ndim ; j++){
	    param->p[i][j] = param->p[0][j];
	}
    }
    for ( i = 1 , j = 0 ; i < param->npts ; i++ , j++){
	param->p[i][j] += side;
    }

    simplex_calc_psum(param);

    for ( i = 0 ; i < param->npts ; i++ ){
	param->y[i] = (*param->func)( param->p[i], param->funcpars );
    }
    param->nfunc += param->npts;

    return param;
}

/*-----------------------simplex_init_simplex-------------------
Effect: Copies an initial simplex p and the function values y at
        its vertices into param->p and param->y.  Evaluates
	param->psum.  Returns param.

Requires:
        param: pointer to Simplex_param.

	p: array of ndim+1 arrays of ndim doubles which are the
	   coordinates of the vertices of the initial simplex.

	y: array of ndim+1 doubles which are the value of
	   *param->func at the points p[i] with parameters
	   param->funcpars.

Modifies: param.

Implementation notes:
---------------------------------------------------------------*/
Simplex_param *
simplex_init_simplex(
	     Simplex_param *param,
	     double         **p,
	     double         y[]
){
    size_t i;
    size_t j;

    for ( j = 0 ; j < param->npts ; j++ ){
	for ( i = 0 ; i < param->ndim ; i++ ){
	    param->p[j][i] = p[j][i];
	}
	param->y[j] = y[j];
    }

    simplex_calc_psum(param);

    return param;
}

/*-------------------------simplex_free--------------------------
Effect: Frees all allocated storage associated with a
        Simplex_param.

Requires: param: pointer to the param to free.

Modifies: param.

Implementation notes:
---------------------------------------------------------------*/
void
simplex_free(Simplex_param *param){
    free(param->ptry);
    free(param->psum);
    free_arrdoub2(param->p);
    free(param->y);
    free(param);
    return;
}
